import { Edge, GraphQueryResult, Node, useML, useSearchResultData } from '@graphpolaris/shared/lib/data-access';
import { Cull } from '@pixi-essentials/cull';
import { dataColors, tailwindColors } from 'config';
import * as d3 from 'd3';
import { Viewport } from 'pixi-viewport';
import {
  Application,
  ColorSource,
  Container,
  DisplayObject,
  FederatedPointerEvent,
  Graphics,
  IPointData,
  Point,
  Sprite,
  Text,
  Texture,
} from 'pixi.js';
import { useEffect, useRef, useState } from 'react';
import { LinkType, NodeType } from '../Types';
import { NLPopup } from './MatrixPopup';

import { Actions, Interpolations } from 'pixi-actions';

import Color from 'color';
import { localConfigSchemaType } from '../../../Types';
import { createColumn } from './ColumnGraphicsComponent';
import { ReorderingManager } from './ReorderingManager';
import { TextLabel } from './TextLabel';

type Props = {
  // onClick: (node: NodeType, pos: IPointData) => void;
  // onHover: (data: { node: NodeType; pos: IPointData }) => void;
  // onUnHover: (data: { node: NodeType; pos: IPointData }) => void;
  highlightNodes: NodeType[];
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  graph?: GraphQueryResult;
  localConfig: localConfigSchemaType;
};

const app = new Application({ background: 0xffffff, antialias: true, autoDensity: true, eventMode: 'auto' });
//////////////////
// MAIN COMPONENT
//////////////////
export const MatrixPixi = (props: Props) => {
  const columnLayer = new Container();
  const columnLabelLayer = new Container();
  const rowsLabelLayer = new Container();

  let lastZoom = Infinity;

  let config = {
    textOffsetX: 100,
    textOffsetY: 100,
    visMapping: [] as any[], // TODO type

    CELL_HEIGHT: 50,
    CELL_WIDTH: 50,

    LABEL_FONT_FAMILY: 'Roboto',
    LABEL_FILL: 0x000000,
    LABEL_ALIGN: 'center',
    LABEL_FONT_SIZE: 25,
    LABEL_MARGIN: 4,
    LABEL_MIN_WIDTH: 6,

    ANIMATION_DURATION: 0.8,
  };

  let columnOrder: string[] = [];
  let rowOrder: string[] = [];
  const [columnOrderState, setColumnOrderState] = useState<string[]>([]);
  const [rowOrderState, setRowOrderState] = useState<string[]>([]);

  const [quickPopup, setQuickPopup] = useState<{ node: NodeType; pos: IPointData } | undefined>();
  const [popups, setPopups] = useState<{ node: NodeType; pos: IPointData }[]>([]);

  const nodeMap = useRef(new Map<string, Graphics>());
  const linkMap = useRef(new Map<string, Graphics>());
  const viewport = useRef<Viewport>();
  const ref = useRef<HTMLDivElement>(null);
  const isSetup = useRef(false);
  const ml = useML();
  const searchResults = useSearchResultData();

  const cull = new Cull();
  let cullDirty = useRef(true);

  function resize() {
    const width = ref?.current?.clientWidth || 1000;
    const height = ref?.current?.clientHeight || 1000;
    app.renderer.resize(width, height);
    if (viewport.current) {
      viewport.current.screenWidth = width;
      viewport.current.worldWidth = width;
      viewport.current.worldHeight = height;
      viewport.current.screenHeight = height;
    }

    if (props.graph) {
      setup();
    }

    app.render();
  }

  useEffect(() => {
    if (!ref.current) return;
    const resizeObserver = new ResizeObserver(() => {
      resize();
    });
    resizeObserver.observe(ref.current);
    return () => resizeObserver.disconnect(); // clean up
  }, []);

  useEffect(() => {
    if (ref.current && ref.current.children.length === 0) {
      ref.current.appendChild(app.view as HTMLCanvasElement);
      resize();
    }
  }, [ref]);

  // useEffect(() => {
  //   console.log('newColumnOrder', columnOrder);
  // }, [columnOrder]);

  useEffect(() => {
    // console.log('graph changed', props.graph, ref.current, ref.current.children.length > 0, imperative.current);
    if (props.graph && ref.current && ref.current.children.length > 0) {
      if (!isSetup.current) setup();
      else update();
    }
  }, [props.graph]);

  useEffect(() => {
    // console.log('graph changed', props.graph, ref.current, ref.current.children.length > 0, imperative.current);
    if (props.graph && ref.current && ref.current.children.length > 0) {
      setup();
    }
  }, [props.localConfig]);

  // TODO implement search results
  // useEffect(() => {
  //   if (props.graph) {
  //     props.graph.nodes.forEach((node: NodeType) => {
  //       const gfx = nodeMap.current.get(node.id);
  //       if (!gfx) return;
  //       const isNodeInSearchResults = searchResults.nodes.some((resultNode) => resultNode.id === node.id);
  //       gfx.alpha = isNodeInSearchResults || searchResults.nodes.length === 0 ? 1 : 0.05;
  //     });

  //     props.graph.links.forEach((link: LinkType) => {
  //       const gfx = linkMap.current.get(link.id);
  //       if (!gfx) return;
  //       const isLinkInSearchResults = searchResults.edges.some((resultEdge) => resultEdge.id === link.id);
  //       gfx.alpha = isLinkInSearchResults || searchResults.edges.length === 0 ? 1 : 0.05;
  //     });
  //   }
  // }, [searchResults]);

  function onButtonDown(event: FederatedPointerEvent) {
    console.log(
      event.currentTarget
      // graph.nodes.find((node) => node.id === event.currentTarget.name)
    );
  }

  const update = (forceClear = false) => {
    setPopups([]);
    if (!props.graph || !ref.current) return;

    if (props.graph) {
      if (forceClear) {
        nodeMap.current.clear();
        linkMap.current.clear();
        rowsLabelLayer.removeChildren();
        columnLayer.removeChildren();
      }
    }
  };

  const onHoverColumn = (event: any, currentNode: Node | null) => {
    app.renderer.events.cursorStyles.pointer = 'crosshair';
    if (!currentNode) {
      rowsLabelLayer.children.forEach((row) => {
        row.alpha = 1;
      });
      columnLayer.children.forEach((col) => {
        col.alpha = 1;
      });
      columnLabelLayer.children.forEach((colText) => {
        colText.alpha = 1;
      });

      return;
    }

    columnLayer.children.forEach((col) => {
      col.alpha = 0.25;
    });
    event.currentTarget.alpha = 1;

    columnLabelLayer.children.forEach((colText) => {
      colText.alpha = 0.25;
      if (colText.name?.endsWith(currentNode.id)) colText.alpha = 1;
    });

    if (!props.graph) return;
    rowsLabelLayer.children.forEach((col) => {
      col.alpha = 0.25;
    });

    if (!props.graph) return;
    rowsLabelLayer.children.forEach((col) => {
      col.alpha = 0.25;
    });

    // console.log(event.currentTarget);
    const edgesForThisColumn = props.graph.edges.filter((edge) => {
      return edge.from === currentNode.id || edge.to === currentNode.id;
    });

    edgesForThisColumn.forEach((edge) => {
      let row = rowsLabelLayer.getChildByName(edge.to);
      if (row) row.alpha = 1;

      row = rowsLabelLayer.getChildByName(edge.from);
      if (row) row.alpha = 1;
    });
  };

  const reorderColumns = (columnOrder: string[]) => {
    const columns = columnLayer.children;
    const columnLabels = columnLabelLayer.children;
    let columnPositions: Point[] = [];
    for (let i = 0; i < columns.length; i++) {
      const col = columns[i];
      columnPositions.push(new Point(col.position.x, col.position.y));
    }

    let columnTextPositions: Point[] = [];
    for (let i = 0; i < columnLabels.length; i++) {
      const col = columnLabels[i];
      columnTextPositions.push(new Point(col.position.x, col.position.y));
    }

    if (columnTextPositions.length !== columnPositions.length)
      throw new Error(
        'columnTextPositions and columnPositions have different length ' + columnTextPositions.length + ' / ' + columnPositions.length
      );

    for (let i = 0; i < columns.length; i++) {
      const column = columns[i];
      if (!column.name) throw new Error('column has no name');
      const index = columnOrder.indexOf(column.name);
      if (index === -1) throw new Error('column not found in columnOrder');

      const newPosition = columnPositions[index];
      // move column to new position
      Actions.moveTo(columns[i], newPosition.x, newPosition.y, 1, Interpolations.pow2out).play();

      const newPositionText = columnTextPositions[index];
      // move column to new position
      Actions.moveTo(columnLabels[i], newPositionText.x, newPositionText.y, 1, Interpolations.pow2out).play();
    }
  };

  const reorderRows = (
    nodesCol: Node[],
    nodesRow: Node[],
    edges: Edge[],
    rowOrder: string[],
    colorScale: any,
    cellWidth: number,
    cellHeight: number
  ) => {
    const rows = rowsLabelLayer.children;
    let rowPositions: Point[] = [];
    for (let i = 0; i < rows.length; i++) {
      const row = rows[i];
      rowPositions.push(new Point(row.position.x, row.position.y));
    }

    for (let i = 0; i < rows.length; i++) {
      const row = rows[i] as DisplayObject;
      if (!row.name) throw new Error('row has no name');
      const index = rowOrder.indexOf(row.name);
      if (index === -1) throw new Error('row not found in rowOrder');

      const newPosition = rowPositions[index];
      // move column to new position
      Actions.moveTo(row, newPosition.x, newPosition.y, 1, Interpolations.pow2out).play();
    }

    const colOrder = columnLayer.children.map((col) => col.name) as string[];
    console.log('colOrder', colOrder);
    if (!colOrder) throw new Error('colOrder is undefined');
    columnLayer.removeChildren();
    setupColumns(edges, colOrder, rowOrder);
  };

  const setup = () => {
    if (!props.graph) throw Error('Graph is undefined; setup not possible');
    console.log('setup matrixvis with graph:', props.graph);

    columnLayer.removeChildren();
    rowsLabelLayer.removeChildren();
    columnLabelLayer.removeChildren();
    viewport.current?.removeAllListeners();
    viewport.current?.removeChildren();
    app.stage.removeChildren();

    const size = ref.current?.getBoundingClientRect();
    viewport.current = new Viewport({
      screenWidth: size?.width || 1000,
      screenHeight: size?.height || 1000,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });

    // setupZoomLogic();

    viewport.current.addChild(columnLayer);

    const groupByType = props.graph.nodes.reduce((group: any, node: Node) => {
      if (!group[node.label]) group[node.label] = [];
      group[node.label].push(node);
      return group;
    }, {} as { [key: string]: Node[] });

    // order groupByType by size
    const ordered = Object.entries(groupByType).sort((a: any[], b: any[]) => b[1].length - a[1].length);

    // console.log('ordered', ordered);
    // console.log('edges', props.graph.edges);

    let cols = [] as Node[];
    let rows = [] as Node[];
    if (ordered.length == 2) {
      cols = ordered[0][1] as Node[];
      rows = ordered[1][1] as Node[];
    } else if (ordered.length == 1) {
      cols = ordered[0][1] as Node[];
      rows = ordered[0][1] as Node[];
    } else {
      // show text that there are no nodes on the viewport
      const finalTextSize = 25;
      const bigRandomScaleFactor = 10;
      const basicText = new Text('Result is empty', { fontSize: finalTextSize * bigRandomScaleFactor, fill: 0x000000, align: 'center' });
      basicText.position.set(viewport.current?.screenWidth / 2 || 1000, viewport.current?.screenHeight / 2 || 1000);
      basicText.scale.set(finalTextSize / (bigRandomScaleFactor * bigRandomScaleFactor));
      basicText.eventMode = 'none';
      viewport.current.addChild(basicText);
      return;
    }

    columnOrder = d3.range(0, cols.length).map((d) => cols[d].id);
    rowOrder = d3.range(0, rows.length).map((d) => rows[d].id);
    const reorderingManager = new ReorderingManager(props.graph);
    const newOrdering = reorderingManager.reorderMatrix('count', columnOrder, rowOrder);
    columnOrder = newOrdering.columnOrder;
    rowOrder = newOrdering.rowOrder;

    setupVisualizationEncodingMapping(props.localConfig);

    setupColumns(props.graph.edges, columnOrder, rowOrder);
    setupColumnLabels(columnOrder);
    setupColumnInteractivity(cols);

    setupRowLabels(rows, rowOrder);
    setupRowInteractivity(cols, rows, rowOrder, d3.scaleOrdinal(d3.schemeCategory10));

    // activate plugins
    viewport.current.drag().pinch().wheel().animate({}).decelerate({ friction: 0.75 });
    lastZoom = viewport.current.scale.x;

    app.stage.addChild(viewport.current);
    app.ticker.add(tick);

    // requestRender();
    update();
    isSetup.current = true;

    cull.clear();
    cull.addAll(viewport.current.children);
    cull.addAll(rowsLabelLayer.children);
    cull.addAll(columnLabelLayer.children);
    cull.addAll(columnLayer.children);
    cullDirty.current = true;

    setColumnOrderState([...columnOrder]);
    setRowOrderState([...rowOrder]);

    resetViewport();
  };

  // let renderRequestId: number | undefined = undefined;
  // const requestRender = () => {
  //   if (renderRequestId) {
  //     return;
  //   }
  //   renderRequestId = window.requestAnimationFrame(() => {
  //     app.render();
  //     renderRequestId = undefined;
  //   });
  // };

  const updateVisibility = () => {
    if (!viewport.current) return;

    // culling
    if (viewport.current.dirty || cullDirty.current) {
      cull.cull(app.renderer.screen);
      // console.log(
      //   'culling',
      //   [...viewport.current.children].filter((x) => x.visible === true).length,
      //   [...viewport.current.children].filter((x) => x.visible === false).length,
      //   [...rowsLabelLayer.children].filter((x) => x.visible === true).length,
      //   [...rowsLabelLayer.children].filter((x) => x.visible === false).length,
      //   [...columnLabelLayer.children].filter((x) => x.visible === true).length,
      //   [...columnLabelLayer.children].filter((x) => x.visible === false).length,
      //   [...columnLayer.children].filter((x) => x.visible === true).length,
      //   [...columnLayer.children].filter((x) => x.visible === false).length
      // );

      viewport.current.dirty = false;
      cullDirty.current = false;
    }

    // levels of detail
    // if (viewport.current?.scale.x !== lastZoom) {
    const zoom = viewport.current?.scale.x || 0.0;

    const oneColumn = columnLayer.children[0];
    if (!oneColumn) return;

    const targetLabelWidth = oneColumn.getBounds().width;
    const targetLabelHeight = targetLabelWidth; //cells have same width and height

    [...columnLabelLayer.children]
      .filter((x) => x.visible === true)
      .forEach((columnItem) => {
        if (targetLabelWidth < config.LABEL_MIN_WIDTH) {
          columnItem.visible = false;
        } else {
          const textLabel = columnItem as TextLabel;
          textLabel.visible = true;
        }
      });

    [...rowsLabelLayer.children]
      .filter((x) => x.visible === true)
      .forEach((rowItem) => {
        if (targetLabelWidth < config.LABEL_MIN_WIDTH) {
          rowItem.visible = false;
        } else {
          rowItem.visible = true;
        }
      });

    lastZoom = zoom;
    // }
  };

  // TODO
  // const setupZoomLogic = () => {
  //   const zoomIn = () => {
  //     if (!viewport.current) return
  //     viewport.current.zoom(-WORLD_WIDTH / 10, true);
  //   };
  //   const zoomOut = () => {
  //     if (!viewport.current) return;
  //     viewport.current.zoom(WORLD_WIDTH / 10, true);
  //   };
  const resetViewport = () => {
    if (!viewport.current) return;
    viewport.current.center = new Point(viewport.current.worldWidth / 2, viewport.current.worldHeight / 2);

    // console.log('resetViewPort', viewport.current.center);
    viewport.current.fitWorld(true);
  };
  // }

  const setupVisualizationEncodingMapping = (localConfig: localConfigSchemaType) => {
    if (!props.graph) throw new Error('Graph is undefined; cannot setup matrix');
    const visMapping = []; // TODO type

    let colorNeutralString = dataColors.neutral['5'];
    const colorNeutral = Color(colorNeutralString);

    // make adjacency
    // const adjacenyScale = d3.scaleLinear([dataColors.neutral['5'], tailwindColors.entity.DEFAULT]);
    const adjacenyScale = d3.scaleLinear([colorNeutral, tailwindColors.entity.DEFAULT]);
    visMapping.push({
      attribute: 'adjacency',
      encoding: localConfig.marks,
      colorScale: adjacenyScale,
      renderFunction: function (i: number, color: ColorSource, gfxContext: Graphics) {
        gfxContext.beginFill(color, 1);

        // TODO get
        // gfx.beginFill(colorScale(node.type)));

        gfxContext.lineStyle(2, 0xffffff);
        if (this.encoding === 'rect') {
          gfxContext.drawRect(0, i * config.CELL_HEIGHT, config.CELL_WIDTH, config.CELL_HEIGHT);
        }
        if (this.encoding === 'circle') {
          gfxContext.drawCircle(config.CELL_WIDTH / 2, i * config.CELL_HEIGHT + config.CELL_HEIGHT / 2, config.CELL_WIDTH / 2);
        }
        gfxContext.endFill();

        return gfxContext;
      },
    });

    // make label
    // const labels = props.graph.edges.map((edge: Edge) => {
    //   return edge.label;
    // });
    // const tailwindColorsArray = Object.values(tailwindColors.entity).slice(0, labels.length);
    // const scale = d3.scaleOrdinal(tailwindColorsArray.reverse());
    // scale.domain(Array.from(labels));
    // console.log('labels', labels, props.graph.edges);
    // visMapping.push({ attribute: 'label', encoding: 'rect', colorScale: scale });

    // make value
    // const values = props.graph.edges.map((edge: Edge) => {
    //   return edge?.attributes?.Distance as number;
    // });

    // const extent = d3.extent(values);
    // if (extent[0] === undefined || extent[1] === undefined) throw new Error('Extent is undefined');
    // console.log('labels', values, extent, props.graph.edges);

    // const scale = d3.scaleLinear(['red', 'blue']);
    // scale.domain(extent as [number, number]);
    // visMapping.push({ attribute: 'attributes.Distance', encoding: 'rect', colorScale: scale });

    config.visMapping = visMapping;
  };

  function setupColumns(edges: Edge[], columnOrder: string[], rowOrder: string[]) {
    const visMapping = config.visMapping[0];

    if (!visMapping) throw new Error('Cannot setup matrix without visMapping');

    for (let j = 0; j < columnOrder.length; j++) {
      const oneColumn = new Container();
      oneColumn.name = columnOrder[j];

      const edgesForThisColumn = edges.filter((edge) => {
        return edge.from === oneColumn.name || edge.to === oneColumn.name;
      });
      // console.log('edgesForThisColumn', oneColumn.name, edgesForThisColumn);

      const col = createColumn(rowOrder, edgesForThisColumn, config.visMapping, config.CELL_WIDTH, config.CELL_HEIGHT);
      oneColumn.addChild(col);
      oneColumn.position.set(j * config.CELL_WIDTH + config.textOffsetX, config.textOffsetY);

      columnLayer.addChild(oneColumn);

      if (columnOrder.length < 100) {
        oneColumn.alpha = 0;
        Actions.sequence(
          Actions.delay(j * 0.005 * Math.random()),
          Actions.fadeIn(oneColumn, config.ANIMATION_DURATION, Interpolations.pow2out)
        ).play();
      } else {
        oneColumn.alpha = 1;
      }
    }
  }

  const setupColumnInteractivity = (cols: Node[]) => {
    for (let j = 0; j < columnLayer.children.length; j++) {
      const oneColumn = columnLayer.children[j];
      oneColumn.eventMode = 'dynamic';
      oneColumn.on('pointerdown', onButtonDown);
      oneColumn.on('pointerover', (event) => {
        const col = cols.find((col) => col.id === oneColumn.name);
        if (col) onHoverColumn(event, col);
      });
      oneColumn.on('pointerout', (event) => {
        onHoverColumn(event, null as any);
      });
    }
  };

  const setupColumnLabels = (columnOrder: string[]) => {
    columnLabelLayer.removeChildren();

    columnLabelLayer.position.set(config.textOffsetX, 0);
    for (let j = 0; j < columnOrder.length; j++) {
      const textLabel = new TextLabel(columnOrder[j], config);
      // const textLabel = new BitmapTextLabel(columnOrder[j]);

      // console.log('font resolved', columnOrder[j]);
      textLabel.position.set(j * config.CELL_WIDTH + config.CELL_WIDTH / 4, config.textOffsetY - 5);
      columnLabelLayer.addChild(textLabel);
    }
    viewport.current?.addChild(columnLabelLayer);

    const columnAxis = new Sprite(Texture.WHITE);
    columnAxis.width = columnOrder.length * config.CELL_WIDTH;
    columnAxis.height = config.textOffsetY;
    columnAxis.tint = 0xaaaaaa;
    columnAxis.position.set(config.textOffsetX, 0);
    columnAxis.alpha = 0.0;
    // columnAxis.interactive = true;
    columnAxis.eventMode = 'dynamic';
    columnAxis.on('pointerdown', (event) => {
      // const reordering = new ReorderColumnsAction(columnOrder);

      if (!props.graph) throw new Error('Graph is undefined; cannot reorder matrix');

      const reorderingManager = new ReorderingManager(props.graph);
      const newOrdering = reorderingManager.reorderMatrix('count', columnOrder, rowOrder);
      columnOrder = newOrdering.columnOrder.reverse();

      reorderColumns(columnOrder);
    });
    // columnAxis.on('pointerdown', reorderColumns);
    columnAxis.on('pointerover', (event) => {
      Actions.fadeTo(columnAxis, 0.5, config.ANIMATION_DURATION, Interpolations.pow2out).play();
    });
    columnAxis.on('pointerout', (event) => {
      Actions.fadeTo(columnAxis, 0.0, config.ANIMATION_DURATION, Interpolations.pow2out).play();
    });
    viewport.current?.addChild(columnAxis);
  };

  function shuffle(array: any[]) {
    var m = array.length,
      t,
      i;

    // While there remain elements to shuffle…
    while (m) {
      // Pick a remaining element…
      i = Math.floor(Math.random() * m--);

      // And swap it with the current element.
      t = array[m];
      array[m] = array[i];
      array[i] = t;
    }

    return array;
  }

  function setupRowLabels(rows: Node[], rowOrder: string[]) {
    rowsLabelLayer.removeChildren();

    // columnLabelLayer.position.set(config.textOffsetX, 0);
    for (let i = 0; i < rowOrder.length; i++) {
      const rowID = rowOrder[i];
      const textLabel = new TextLabel(rowID, config);
      textLabel.anchor.x = 1;
      textLabel.anchor.y = 0;
      textLabel.rotation = 0;
      textLabel.name = rowID;
      textLabel.position.set(config.textOffsetX - 5, i * config.CELL_HEIGHT + config.textOffsetY + config.CELL_HEIGHT / 4);
      rowsLabelLayer.addChild(textLabel);

      if (rowOrder.length < 100) {
        textLabel.alpha = 0;
        Actions.sequence(
          Actions.delay(i * 0.005 * Math.random()),
          Actions.fadeIn(textLabel, config.ANIMATION_DURATION, Interpolations.pow2out)
        ).play();
      } else {
        textLabel.alpha = 1;
      }
    }

    viewport.current?.addChild(rowsLabelLayer);
  }

  const setupRowInteractivity = (cols: Node[], rows: Node[], rowOrder: string[], colorScale: any) => {
    const rowAxis = new Sprite(Texture.WHITE);
    rowAxis.width = config.textOffsetX;
    rowAxis.height = rows.length * config.CELL_HEIGHT;
    rowAxis.tint = 0xaaaaaa;
    rowAxis.position.set(0, config.textOffsetY);
    rowAxis.alpha = 0.0;
    rowAxis.eventMode = 'dynamic';
    rowAxis.on('pointerdown', (event) => {
      if (props.graph) {
        rowOrder = shuffle(rowOrder);
        reorderRows(cols, rows, props?.graph.edges, rowOrder, colorScale, config.CELL_WIDTH, config.CELL_HEIGHT);
      }
    });
    rowAxis.on('pointerover', (event) => {
      Actions.fadeTo(rowAxis, 0.5, 0.5, Interpolations.pow2out).play();
    });
    rowAxis.on('pointerout', (event) => {
      Actions.fadeTo(rowAxis, 0.0, 0.5, Interpolations.pow2out).play();
    });
    viewport.current?.addChild(rowAxis);
  };

  const tick = (delta: number) => {
    if (props.graph) {
      Actions.tick(delta / 60);

      updateVisibility();
    }
  };

  return (
    <>
      {popups.map((popup) => (
        <NLPopup onClose={() => {}} data={popup} key={popup.node.id} />
      ))}
      {quickPopup && <NLPopup onClose={() => {}} data={quickPopup} />}

      {/* {columnOrderState.length > 0 && <ColumnLabelTrack columnLabels={columnOrderState} config={config} />} */}
      <div className="h-full w-full overflow-hidden" ref={ref}></div>
    </>
  );
};
